package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** a tree (or node in a tree); see jargon.txt for details */
public class Tree<T>
    extends PrintableTree<Tree<T>>
    implements Iterable<Tree<T>>,
               GraphViz.ToGraphViz {

    final T           head;
          Tree<T>[]   children;
          Object[]    labels;
    final Input.Location    location;

    public T                 head()        { return head; }
    public int               numChildren() { return children.length; }
    public Iterable<Tree<T>> children()    { return new ArrayIterator(children); }
    public Iterator<Tree<T>> iterator()    { return new ArrayIterator(children); }
    public Tree<T>           child(int i)  { return children[i]; }
    public Object            label(int i)  { return labels[i]; }

    public Input.Location    getLocation() { return location; }

    public Tree(Input.Location loc, T head)                   { this(loc, head, null); }
    public Tree(Input.Location loc, T head, Tree<T>[] children) { this(loc, head, children, null); }
    public Tree(Input.Location loc, T head, Tree<T>[] children, Object[] labels) {
        this.location = loc;
        this.head = head;

        Tree<T>[] children2 = children==null ? new Tree[0] : new Tree[children.length];
        if (children != null) System.arraycopy(children, 0, children2, 0, children.length);
        this.children = children2;

        Object[] labels2 = labels==null ? new Object[0] : new Object[labels.length];
        if (labels != null) System.arraycopy(labels, 0, labels2, 0, labels.length);
        this.labels = labels2;
    }

    protected String headToString() { return head==null?null:head.toString(); }
    protected String headToJava()   {
        if (head==null) return null;
        if (head instanceof ToJava) {
            StringBuffer sb = new StringBuffer();
            ((ToJava)head).toJava(sb);
            return sb.toString();
        }
        return head==null?"null":("\""+StringUtil.toJavaString(head.toString())+"\"");
    }
    protected String left()   { return "{"; }
    protected String right()  { return "}"; }
    protected boolean ignoreSingleton() { return false; }

    public GraphViz.Node toGraphViz(GraphViz gv) {
        if (gv.hasNode(this)) return gv.createNode(this);
        GraphViz.Node n = gv.createNode(this);
        n.label = head()==null ? "" : head().toString();
        //n.color = "red";
        for(Tree t : this) n.edge(t, null);
        return n;
    }
    public boolean isTransparent() { return false; }
    public boolean isHidden() { return false; }

    public static interface TreeFunctor<T,R> extends Functor<Iterable<Tree<T>>, R> {
    }

    public static class ArrayBuildingTreeFunctor<T> implements TreeFunctor<T,T[]>, ToJava {

        public void toJava(StringBuffer sb) { sb.append("new Tree.ArrayBuildingTreeFunctor()"); }
        public String toString() { return ""; }
        
        public T[] invoke(Iterable<Tree<T>> t) {
            ArrayList ret = new ArrayList();
            for(Tree tc : t) {
                if (tc.head() != null && tc.head() instanceof Functor)
                    ret.add(((Functor<Iterable<Tree>,Object>)tc.head()).invoke(tc.children()));
                else if (tc.numChildren() == 0)
                    ret.add(tc.head());
                else {
                    System.err.println("FIXME: don't know what to do about " + tc);
                    ret.add(null);
                }
            }
            return (T[])ret.toArray(new Object[0]);
        }
    }
}
